@@ -25,6 +25,7 @@
 #include <compat.h>
 
 #include <sys/types.h>
+#include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 
@@ -51,7 +52,7 @@
 #define PROGNAME "update-alternatives"
 
 static const char *altdir = SYSCONFDIR "/alternatives";
-static const char *admdir;
+static char *admdir = NULL;
 static const char *instdir = "";
 static size_t instdir_len;
 
@@ -101,7 +102,7 @@ enum output_mode {
 
 /* Action to perform */
 static enum action action = ACTION_NONE;
-static const char *log_file = LOGDIR "/alternatives.log";
+static char *log_file = NULL;
 static FILE *fh_log = NULL;
 /* Skip alternatives properly configured in auto mode (for --config) */
 static int opt_skip_auto = 0;
@@ -162,8 +163,10 @@ usage(void)
 
 	printf(_(
 "Options:\n"
-"  --altdir <directory>     change the alternatives directory.\n"
-"  --admindir <directory>   change the administrative directory.\n"
+"  --altdir <directory>     change the alternatives directory\n"
+"                             (default is %s).\n"
+"  --admindir <directory>   change the administrative directory\n"
+"                             (default is %s).\n"
 "  --instdir <directory>    change the installation directory.\n"
 "  --root <directory>       change the filesystem root directory.\n"
 "  --log <file>             change the log file.\n"
@@ -175,7 +178,7 @@ usage(void)
 "  --debug                  debug output, way more output.\n"
 "  --help                   show this help message.\n"
 "  --version                show the version.\n"
-));
+), altdir, admdir);
 }
 
 static void DPKG_ATTR_NORET DPKG_ATTR_PRINTF(1)
@@ -519,10 +522,13 @@ log_msg(const char *fmt, ...)
 	if (fh_log) {
 		char timestamp[64];
 		time_t now;
+		struct tm tm;
 
 		time(&now);
+		if (localtime_r(&now, &tm) == NULL)
+			syserr(_("cannot get local time to log into '%s'"), log_file);
 		strftime(timestamp, sizeof(timestamp), "%Y-%m-%d %H:%M:%S",
-		         localtime(&now));
+		         &tm);
 		fprintf(fh_log, "%s %s: ", PROGNAME, timestamp);
 		va_start(args, fmt);
 		vfprintf(fh_log, fmt, args);
@@ -560,9 +566,9 @@ fsys_set_dir(const char *dir)
 }
 
 static char *
-fsys_gen_admindir(const char *basedir)
+fsys_gen_admindir(void)
 {
-	return xasprintf("%s%s/%s", instdir, basedir, "alternatives");
+	return fsys_get_path(ADMINDIR "/alternatives");
 }
 
 static bool
@@ -626,6 +632,37 @@ fsys_xreadlink(const char *linkname)
 	return buf;
 }
 
+static void
+fsys_set_ref_time(const char *linkname, const char *target)
+{
+#ifdef HAVE_LUTIMES
+	/* If the symlink did not exist, then copy the timestamps
+	 * from the target. This is needed so we can get reproducible
+	 * installations, for programs that track these timestamps on
+	 * their databases. */
+	struct stat st;
+	struct timeval tv[2];
+	char *root_linkname;
+
+	if (fsys_lstat(target, &st) < 0) {
+		if (errno != ENOENT)
+			syserr(_("unable to get file '%s%s' metadata"),
+			       instdir, target);
+		return;
+	}
+
+	tv[0].tv_sec = st.st_mtime;
+	tv[0].tv_usec = 0;
+	tv[1].tv_sec = st.st_mtime;
+	tv[1].tv_usec = 0;
+
+	root_linkname = fsys_get_path(linkname);
+	if (lutimes(root_linkname, tv) < 0 && errno != ENOSYS)
+		syserr(_("cannot set symlink '%s' timestamp"), root_linkname);
+	free(root_linkname);
+#endif
+}
+
 static void
 fsys_symlink(const char *filename, const char *linkname)
 {
@@ -633,6 +670,9 @@ fsys_symlink(const char *filename, const char *linkname)
 
 	root_linkname = fsys_get_path(linkname);
 
+	if (unlink(root_linkname) < 0 && errno != ENOENT)
+		syserr(_("unable to remove '%s'"), root_linkname);
+
 	if (symlink(filename, root_linkname))
 		syserr(_("error creating symbolic link '%.255s'"), root_linkname);
 
@@ -805,6 +845,7 @@ struct commit_operation {
 		OPCODE_NOP,
 		OPCODE_RM,
 		OPCODE_MV,
+		OPCODE_REF_TIME,
 	} opcode;
 
 	char *arg_a;
@@ -1229,6 +1270,14 @@ struct altdb_context {
 	jmp_buf on_error;
 };
 
+static void
+altdb_context_free(struct altdb_context *ctx)
+{
+	if (ctx->fh)
+		fclose(ctx->fh);
+	free(ctx->filename);
+}
+
 static int
 altdb_filter_namelist(const struct dirent *entry)
 {
@@ -1457,16 +1506,16 @@ alternative_load(struct alternative *a, enum altdb_flags flags)
 	/* Open the alternative file. */
 	ctx.fh = fopen(ctx.filename, "r");
 	if (ctx.fh == NULL) {
-		if (errno == ENOENT)
+		if (errno == ENOENT) {
+			altdb_context_free(&ctx);
 			return false;
+		}
 
 		syserr(_("unable to open file '%s'"), ctx.filename);
 	}
 
 	if (setjmp(ctx.on_error)) {
-		if (ctx.fh)
-			fclose(ctx.fh);
-		free(ctx.filename);
+		altdb_context_free(&ctx);
 		alternative_reset(a);
 		return false;
 	}
@@ -1474,8 +1523,11 @@ alternative_load(struct alternative *a, enum altdb_flags flags)
 	/* Verify the alternative is not empty. */
 	if (fstat(fileno(ctx.fh), &st) == -1)
 		syserr(_("cannot stat file '%s'"), ctx.filename);
-	if (st.st_size == 0)
+	if (st.st_size == 0) {
+		altdb_context_free(&ctx);
+		alternative_reset(a);
 		return false;
+	}
 
 	/* Start parsing mandatory attributes (link+status) of the alternative */
 	alternative_reset(a);
@@ -1831,12 +1883,6 @@ alternative_config(struct alternative *a, const char *current_choice)
 		pr(_("Nothing to configure."));
 	} else if (opt_skip_auto && a->status == ALT_ST_AUTO) {
 		alternative_display_user(a);
-	} else if (alternative_choices_count(a) == 1 &&
-	           a->status == ALT_ST_AUTO &&
-	           current_choice != NULL) {
-		pr(_("There is only one alternative in link group %s (providing %s): %s"),
-		   a->master_name, a->master_link, current_choice);
-		pr(_("Nothing to configure."));
 	} else {
 		new_choice = alternative_select_choice(a);
 	}
@@ -1881,6 +1927,9 @@ alternative_commit(struct alternative *a)
 		case OPCODE_MV:
 			fsys_mv(op->arg_a, op->arg_b);
 			break;
+		case OPCODE_REF_TIME:
+			fsys_set_ref_time(op->arg_a, op->arg_b);
+			break;
 		}
 	}
 
@@ -1966,20 +2015,23 @@ alternative_prepare_install_single(struct alternative *a, const char *name,
 		free(root_altdir);
 	}
 
+	fn = xasprintf("%s/%s", altdir, name);
+
 	/* Create link in /etc/alternatives. */
 	fntmp = xasprintf("%s/%s" ALT_TMP_EXT, altdir, name);
-	fn = xasprintf("%s/%s", altdir, name);
-	fsys_rm(fntmp);
 	fsys_symlink(file, fntmp);
 	alternative_add_commit_op(a, OPCODE_MV, fntmp, fn);
+	if (fsys_pathname_is_missing(fn))
+		alternative_add_commit_op(a, OPCODE_REF_TIME, fn, file);
 	free(fntmp);
 
 	if (alternative_path_needs_update(linkname, fn)) {
 		/* Create alternative link. */
 		fntmp = xasprintf("%s" ALT_TMP_EXT, linkname);
-		fsys_rm(fntmp);
 		fsys_symlink(fn, fntmp);
 		alternative_add_commit_op(a, OPCODE_MV, fntmp, linkname);
+		if (fsys_pathname_is_missing(linkname))
+			alternative_add_commit_op(a, OPCODE_REF_TIME, linkname, fn);
 		free(fntmp);
 	}
 	free(fn);
@@ -2052,11 +2104,11 @@ alternative_remove_files(struct alternative *a)
 	xunlink_args("%s/%s", admdir, a->master_name);
 }
 
-static const char *
+static char *
 alternative_remove(struct alternative *a, const char *current_choice,
                    const char *path)
 {
-	const char *new_choice = NULL;
+	char *new_choice = NULL;
 
 	if (alternative_has_choice(a, path))
 		alternative_remove_choice(a, path);
@@ -2077,40 +2129,50 @@ alternative_remove(struct alternative *a, const char *current_choice,
 		}
 		best = alternative_get_best(a);
 		if (best)
-			new_choice = best->master_file;
+			new_choice = xstrdup(best->master_file);
 	}
 
 	return new_choice;
 }
 
+static bool
+alternative_has_broken_symlink(const char *linkname, const char *ref_target)
+{
+	char *target;
+
+	target = fsys_areadlink(linkname);
+	if (!target)
+		return true;
+	if (strcmp(target, ref_target) != 0) {
+		free(target);
+		return true;
+	}
+	free(target);
+	return false;
+}
+
 static bool
 alternative_has_broken_slave(struct slave_link *sl, struct fileset *fs)
 {
 	if (fileset_can_install_slave(fs, sl->name)) {
 		char *wanted;
-		char *sl_altlnk, *sl_current;
+		const char *sl_target;
 
 		/* Verify link -> /etc/alternatives/foo */
-		sl_altlnk = fsys_areadlink(sl->link);
-		if (!sl_altlnk)
-			return true;
 		wanted = xasprintf("%s/%s", altdir, sl->name);
-		if (strcmp(sl_altlnk, wanted) != 0) {
+		if (alternative_has_broken_symlink(sl->link, wanted)) {
 			free(wanted);
-			free(sl_altlnk);
 			return true;
 		}
-		free(sl_altlnk);
+
 		/* Verify /etc/alternatives/foo -> file */
-		sl_current = fsys_areadlink(wanted);
-		free(wanted);
-		if (!sl_current)
-			return true;
-		if (strcmp(sl_current, fileset_get_slave(fs, sl->name)) != 0) {
-			free(sl_current);
+		sl_target = fileset_get_slave(fs, sl->name);
+		if (alternative_has_broken_symlink(wanted, sl_target)) {
+			free(wanted);
 			return true;
 		}
-		free(sl_current);
+
+		free(wanted);
 	} else {
 		char *sl_altlnk;
 
@@ -2133,22 +2195,17 @@ alternative_needs_update(struct alternative *a)
 {
 	enum alternative_update_reason reason = ALT_UPDATE_NO;
 	const char *current;
-	char *altlnk, *wanted;
+	char *wanted;
 	struct fileset *fs;
 	struct slave_link *sl;
 
 	/* Check master link */
-	altlnk = fsys_areadlink(a->master_link);
-	if (!altlnk)
-		return ALT_UPDATE_LINK_BROKEN;
 	wanted = xasprintf("%s/%s", altdir, a->master_name);
-	if (strcmp(altlnk, wanted) != 0) {
+	if (alternative_has_broken_symlink(a->master_link, wanted)) {
 		free(wanted);
-		free(altlnk);
 		return ALT_UPDATE_LINK_BROKEN;
 	}
 	free(wanted);
-	free(altlnk);
 
 	/* Stop if we have an unmanaged alternative */
 	current = alternative_get_current(a);
@@ -2282,13 +2339,13 @@ alternative_map_free(struct alternative_map *am)
 	}
 }
 
-static const char *
+static char *
 alternative_set_manual(struct alternative *a, const char *path)
 {
-	const char *new_choice = NULL;
+	char *new_choice = NULL;
 
 	if (alternative_has_choice(a, path))
-		new_choice = path;
+		new_choice = xstrdup(path);
 	else
 		error(_("alternative %s for %s not registered; "
 		        "not setting"), path, a->master_name);
@@ -2297,17 +2354,17 @@ alternative_set_manual(struct alternative *a, const char *path)
 	return new_choice;
 }
 
-static const char *
+static char *
 alternative_set_auto(struct alternative *a)
 {
-	const char *new_choice = NULL;
+	char *new_choice = NULL;
 
 	alternative_set_status(a, ALT_ST_AUTO);
 	if (alternative_choices_count(a) == 0)
-		pr(_("There is no program which provides %s."),
-		   a->master_name);
+		info(_("there is no program which provides %s"),
+		     a->master_name);
 	else
-		new_choice = alternative_get_best(a)->master_file;
+		new_choice = xstrdup(alternative_get_best(a)->master_file);
 
 	return new_choice;
 }
@@ -2433,6 +2490,35 @@ alternative_evolve(struct alternative *a, struct alternative *b,
 	}
 }
 
+static char *
+alternative_install(struct alternative **aptr, struct alternative *inst_alt,
+                    const char *current_choice, struct fileset *fileset)
+{
+	struct alternative *a = *aptr;
+	char *new_choice = NULL;
+
+	if (a->master_link) {
+		/* Alternative already exists, check if anything got
+		 * updated. */
+		alternative_evolve(a, inst_alt, current_choice, fileset);
+		alternative_free(inst_alt);
+	} else {
+		/* Alternative doesn't exist, create from parameters. */
+		alternative_free(a);
+		*aptr = a = inst_alt;
+	}
+	alternative_add_choice(a, fileset);
+	if (a->status == ALT_ST_AUTO) {
+		new_choice = xstrdup(alternative_get_best(a)->master_file);
+	} else {
+		verbose(_("automatic updates of %s/%s are disabled; "
+		          "leaving it alone"), altdir, a->master_name);
+		verbose(_("to return to automatic updates use "
+		          "'%s --auto %s'"), PROGNAME, a->master_name);
+	}
+	return new_choice;
+}
+
 static void
 alternative_update(struct alternative *a,
                    const char *current_choice, const char *new_choice)
@@ -2554,15 +2640,18 @@ alternative_set_selection(struct alternative_map *all, const char *name,
 	debug("set_selection(%s, %s, %s)", name, status, choice);
 	a = alternative_map_find(all, name);
 	if (a) {
-		const char *new_choice = NULL;
+		char *new_choice = NULL;
 
 		if (strcmp(status, "auto") == 0) {
+			info(_("selecting alternative %s as auto"), name);
 			new_choice = alternative_set_auto(a);
 		} else if (alternative_has_choice(a, choice)) {
+			info(_("selecting alternative %s as choice %s"), name,
+			     choice);
 			new_choice = alternative_set_manual(a, choice);
 		} else {
-			pr(_("Alternative %s unchanged because choice "
-			     "%s is not available."), name, choice);
+			info(_("alternative %s unchanged because choice "
+			       "%s is not available"), name, choice);
 		}
 
 		if (new_choice) {
@@ -2572,9 +2661,11 @@ alternative_set_selection(struct alternative_map *all, const char *name,
 			alternative_select_mode(a, current_choice);
 
 			alternative_update(a, current_choice, new_choice);
+
+			free(new_choice);
 		}
 	} else {
-		pr(_("Skip unknown alternative %s."), name);
+		info(_("skip unknown alternative %s"), name);
 	}
 }
 
@@ -2613,8 +2704,7 @@ alternative_set_selections(FILE *input, const char *desc)
 		while (i < len && !isblank(line[i]))
 			i++;
 		if (i >= len) {
-			printf("[%s %s] ", PROGNAME, "--set-selections");
-			pr(_("Skip invalid line: %s"), line);
+			info(_("skip invalid selection line: %s"), line);
 			continue;
 		}
 		line[i++] = '\0';
@@ -2626,8 +2716,7 @@ alternative_set_selections(FILE *input, const char *desc)
 		while (i < len && !isblank(line[i]))
 			i++;
 		if (i >= len) {
-			printf("[%s %s] ", PROGNAME, "--set-selections");
-			pr(_("Skip invalid line: %s"), line);
+			info(_("skip invalid selection line: %s"), line);
 			continue;
 		}
 		line[i++] = '\0';
@@ -2636,13 +2725,11 @@ alternative_set_selections(FILE *input, const char *desc)
 
 		/* Delimit choice string in the line */
 		if (i >= len) {
-			printf("[%s %s] ", PROGNAME, "--set-selections");
-			pr(_("Skip invalid line: %s"), line);
+			info(_("skip invalid selection line: %s"), line);
 			continue;
 		}
 		choice = line + i;
 
-		printf("[%s %s] ", PROGNAME, "--set-selections");
 		alternative_set_selection(alt_map_obj, name, status, choice);
 	}
 
@@ -2793,27 +2880,27 @@ static const char *
 set_rootdir(const char *dir)
 {
 	instdir = fsys_set_dir(dir);
+	free(log_file);
 	log_file = fsys_get_path(LOGDIR "/alternatives.log");
 	altdir = SYSCONFDIR "/alternatives";
-	admdir = fsys_gen_admindir(dir);
+	free(admdir);
+	admdir = fsys_gen_admindir();
 
 	return instdir;
 }
 
-static const char *
+static char *
 admindir_init(void)
 {
-	const char *basedir, *basedir_env;
+	const char *basedir_env;
 
 	/* Try to get the admindir from an environment variable, usually set
 	 * by the system package manager. */
 	basedir_env = getenv(ADMINDIR_ENVVAR);
 	if (basedir_env)
-		basedir = basedir_env;
+		return xasprintf("%s%s", basedir_env, "/alternatives");
 	else
-		basedir = ADMINDIR;
-
-	return fsys_gen_admindir(basedir);
+		return fsys_gen_admindir();
 }
 
 #define MISSING_ARGS(nb) (argc < i + nb + 1)
@@ -2828,9 +2915,9 @@ main(int argc, char **argv)
 	/* Set of files to install in the alternative. */
 	struct fileset *fileset = NULL;
 	/* Path of alternative we are offering. */
-	char *path = NULL;
+	const char *path = NULL;
 	const char *current_choice = NULL;
-	const char *new_choice = NULL;
+	char *new_choice = NULL;
 	bool modifies_alt = false;
 	bool modifies_sys = false;
 	int i = 0;
@@ -2843,6 +2930,7 @@ main(int argc, char **argv)
 
 	instdir = fsys_set_dir(NULL);
 	admdir = admindir_init();
+	log_file = fsys_get_path(LOGDIR "/alternatives.log");
 
 	if (setvbuf(stdout, NULL, _IONBF, 0))
 		syserr("setvbuf failed");
@@ -2865,6 +2953,7 @@ main(int argc, char **argv)
 		} else if (strcmp("--debug", argv[i]) == 0) {
 			opt_verbose = OUTPUT_DEBUG;
 		} else if (strcmp("--install", argv[i]) == 0) {
+			const char *alink, *aname, *apath;
 			char *prio_str, *prio_end;
 			long prio;
 
@@ -2873,22 +2962,28 @@ main(int argc, char **argv)
 				badusage(_("--%s needs <link> <name> <path> "
 					   "<priority>"), argv[i] + 2);
 
+			alink = argv[i + 1];
+			aname = argv[i + 2];
+			apath = argv[i + 3];
 			prio_str = argv[i + 4];
 
-			if (strcmp(argv[i+1], argv[i+3]) == 0)
-				badusage(_("<link> and <path> can't be the same"));
+			if (strcmp(alink, apath) == 0)
+				badusage(_("<link> '%s' is the same as <path>"),
+				         alink);
 			errno = 0;
 			prio = strtol(prio_str, &prio_end, 10);
 			if (prio_str == prio_end || *prio_end != '\0')
-				badusage(_("priority must be an integer"));
+				badusage(_("priority '%s' must be an integer"),
+				         prio_str);
 			if (prio < INT_MIN || prio > INT_MAX || errno == ERANGE)
-				badusage(_("priority is out of range"));
+				badusage(_("priority '%s' is out of range"),
+				         prio_str);
 
-			a = alternative_new(argv[i + 2]);
-			inst_alt = alternative_new(argv[i + 2]);
+			a = alternative_new(aname);
+			inst_alt = alternative_new(aname);
 			alternative_set_status(inst_alt, ALT_ST_AUTO);
-			alternative_set_link(inst_alt, argv[i + 1]);
-			fileset = fileset_new(argv[i + 3], prio);
+			alternative_set_link(inst_alt, alink);
+			fileset = fileset_new(apath, prio);
 
 			i += 4;
 		} else if (strcmp("--remove", argv[i]) == 0 ||
@@ -2898,7 +2993,7 @@ main(int argc, char **argv)
 				badusage(_("--%s needs <name> <path>"), argv[i] + 2);
 
 			a = alternative_new(argv[i + 1]);
-			path = xstrdup(argv[i + 2]);
+			path = argv[i + 2];
 
 			alternative_check_name(a->master_name);
 			alternative_check_path(path);
@@ -2938,22 +3033,23 @@ main(int argc, char **argv)
 			spath = argv[i + 3];
 
 			if (strcmp(slink, spath) == 0)
-				badusage(_("<link> and <path> can't be the same"));
+				badusage(_("<link> '%s' is the same as <path>"),
+				         slink);
 			if (strcmp(inst_alt->master_name, sname) == 0)
-				badusage(_("name %s is both primary and slave"),
+				badusage(_("<name> '%s' is both primary and slave"),
 				         sname);
 			if (strcmp(slink, inst_alt->master_link) == 0)
-				badusage(_("link %s is both primary and slave"),
+				badusage(_("<link> '%s' is both primary and slave"),
 				         slink);
 			if (alternative_has_slave(inst_alt, sname))
-				badusage(_("duplicate slave name %s"), sname);
+				badusage(_("duplicate slave <name> '%s'"), sname);
 
 			for (sl = inst_alt->slaves; sl; sl = sl->next) {
 				const char *linkname = sl->link;
 				if (linkname == NULL)
 					linkname = "";
 				if (strcmp(linkname, slink) == 0)
-					badusage(_("duplicate slave link %s"),
+					badusage(_("duplicate slave <link> '%s'"),
 					          slink);
 			}
 
@@ -2965,6 +3061,7 @@ main(int argc, char **argv)
 			if (MISSING_ARGS(1))
 				badusage(_("--%s needs a <file> argument"),
 				         argv[i] + 2);
+			free(log_file);
 			log_file = fsys_get_path(argv[i + 1]);
 			i++;
 		} else if (strcmp("--altdir", argv[i]) == 0) {
@@ -2975,14 +3072,15 @@ main(int argc, char **argv)
 			i++;
 
 			/* If altdir is below instdir, convert it to a relative
-			 * path, as we will prepenr instdir as needed. */
+			 * path, as we will prepend instdir as needed. */
 			if (strncmp(altdir, instdir, instdir_len) == 0)
 				altdir += instdir_len;
 		} else if (strcmp("--admindir", argv[i]) == 0) {
 			if (MISSING_ARGS(1))
 				badusage(_("--%s needs a <directory> argument"),
 				         argv[i] + 2);
-			admdir = argv[i + 1];
+			free(admdir);
+			admdir = xstrdup(argv[i + 1]);
 			i++;
 		} else if (strcmp("--instdir", argv[i]) == 0) {
 			if (MISSING_ARGS(1))
@@ -2992,7 +3090,7 @@ main(int argc, char **argv)
 			i++;
 
 			/* If altdir is below instdir, convert it to a relative
-			 * path, as we will prepenr instdir as needed. */
+			 * path, as we will prepend instdir as needed. */
 			if (strncmp(altdir, instdir, instdir_len) == 0)
 				altdir += instdir_len;
 		} else if (strcmp("--root", argv[i]) == 0) {
@@ -3017,6 +3115,8 @@ main(int argc, char **argv)
 		         "config", "set", "set-selections", "install",
 		         "remove", "all", "remove-all", "auto");
 
+	debug("root=%s admdir=%s altdir=%s", instdir, admdir, altdir);
+
 	/* The following actions might modify the current alternative. */
 	if (action == ACTION_SET ||
 	    action == ACTION_AUTO ||
@@ -3046,11 +3146,14 @@ main(int argc, char **argv)
 		if (!alternative_load(a, ALTDB_WARN_PARSER))
 			error(_("no alternatives for %s"), a->master_name);
 	} else if (action == ACTION_REMOVE) {
-		/* FIXME: Be consistent for now with the case when we
+		/* XXX: Be consistent for now with the case when we
 		 * try to remove a non-existing path from an existing
 		 * link group file. */
 		if (!alternative_load(a, ALTDB_WARN_PARSER)) {
 			verbose(_("no alternatives for %s"), a->master_name);
+			alternative_free(a);
+			free(log_file);
+			free(admdir);
 			exit(0);
 		}
 	} else if (action == ACTION_INSTALL) {
@@ -3090,28 +3193,18 @@ main(int argc, char **argv)
 	} else if (action == ACTION_REMOVE_ALL) {
 		alternative_choices_free(a);
 	} else if (action == ACTION_INSTALL) {
-		if (a->master_link) {
-			/* Alternative already exists, check if anything got
-			 * updated. */
-			alternative_evolve(a, inst_alt, current_choice, fileset);
-		} else {
-			/* Alternative doesn't exist, create from parameters. */
-			alternative_free(a);
-			a = inst_alt;
-		}
-		alternative_add_choice(a, fileset);
-		if (a->status == ALT_ST_AUTO) {
-			new_choice = alternative_get_best(a)->master_file;
-		} else {
-			verbose(_("automatic updates of %s/%s are disabled; "
-			          "leaving it alone"), altdir, a->master_name);
-			verbose(_("to return to automatic updates use "
-			          "'%s --auto %s'"), PROGNAME, a->master_name);
-		}
+		new_choice = alternative_install(&a, inst_alt, current_choice,
+		                                 fileset);
 	}
 
 	if (modifies_alt)
 		alternative_update(a, current_choice, new_choice);
 
+	if (a)
+		alternative_free(a);
+	free(new_choice);
+	free(log_file);
+	free(admdir);
+
 	return 0;
 }
